//-< ClassDescriptor.java >------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 13-Jan-99    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Class reponsible for packing/unpacking persistent objects
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
    using System.Reflection;
		
    public class ClassDescriptor
    {
        //
        // GOODS specific ClassDescriptor fields
        //
		
        internal static System.Collections.Hashtable classDictionary = new System.Collections.Hashtable();
		
        internal String name;
        internal int fixedSize;
        internal int varyingSize;
        internal int nFixedReferences;
        internal int nVaryingReferences;
        internal int nFields;
        internal int totalNamesSize;
        internal int cid;
		
        internal static int classCount;
		
        internal ClassDescriptor newClass;
		
        internal Type appClass;
        internal byte[] dbsClass;
		
        internal bool conversionNeeded;
		
        internal ClassDescriptor superclass;
		
        internal FieldDescriptor firstField;
        internal FieldDescriptor lastField;
        internal FieldDescriptor refFields;
        internal FieldDescriptor varyingField;
        internal FieldDescriptor strFields;
        internal FieldDescriptor binFields;
		
        internal System.Reflection.ConstructorInfo constructor;
        internal Object[] constructorParameters;
        internal Metaobject defaultMetaobject;
		
        internal static Type[] constructorProfile = {typeof(Metaobject)};
		
        internal void setDefaultMetaobject()
        {
            Object[] attribs = appClass.GetCustomAttributes(typeof(MetaobjectAttribute), true);
            defaultMetaobject = ((MetaobjectAttribute)attribs[0]).metaobject;
            constructorParameters = new Object[1];
            constructorParameters[0] = defaultMetaobject;
        }

        internal static bool FindTypeByName(Type t, object name) 
        { 
            return t.FullName.Equals(name);
        }

        internal static Type getClass(byte[] desc)
        {
            int nFields = Protocol.unpack4(desc, Protocol.cds_n_fields_offs);
            String name = Protocol.unpackAscii(desc, Protocol.cds_header_size + nFields * Protocol.fds_size);
            return findClass(name);
        }

        internal static Type findClass(String name)
        {
            Type type = null;
            foreach (Assembly ass in AppDomain.CurrentDomain.GetAssemblies()) 
            { 
                foreach (Module mod in ass.GetModules()) 
                { 
                    foreach (Type t in mod.FindTypes(new TypeFilter(FindTypeByName), name)) 
                    { 
                        if (type != null) 
                        { 
                            throw new ApplicationException("Class " + name + " exists in more than one scope");
                        } 
                        else 
                        { 
                            type = t;
                        }
                    }
                }
            }
            if (type == null)
            {
                throw new ApplicationException("Class " + name + " not found in application");
            }
            return type;
        }
		
        internal void  initialize(Type cls)
        {
            appClass = cls;
            name = cls.FullName;
            constructor = appClass.GetConstructor(System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.DeclaredOnly, null, constructorProfile, null);
            if (constructor == null) 
            { 
                throw new ClassDefinitionError("No constructor with metaobject parameter in class " + name);
            }
            setDefaultMetaobject();
        }
		
		
        internal ClassDescriptor(ClassDescriptor newClass, byte[] desc)
        {
            initialize(newClass.appClass);
            this.newClass = newClass;
            fixedSize = Protocol.unpack4(desc, Protocol.cds_fixed_size_offs);
            varyingSize = Protocol.unpack4(desc, Protocol.cds_varying_size_offs);
            nFixedReferences = Protocol.unpack4(desc, Protocol.cds_n_fixed_refs_offs);
            nVaryingReferences = Protocol.unpack4(desc, Protocol.cds_n_varying_refs_offs);
            nFields = Protocol.unpack4(desc, Protocol.cds_n_fields_offs);
            totalNamesSize = Protocol.unpack4(desc, Protocol.cds_total_names_size_offs);
            dbsClass = desc;
            conversionNeeded = true;
            buildFieldsList(desc, Protocol.cds_header_size);
        }
		
		
        internal int buildFieldsList(byte[] desc, int offs)
        {
            int dbsType, nameOffs, size, nItems, next;
            String fieldName;
            int n = 0;
			
            do 
            {
                dbsType = Protocol.unpack2(desc, offs + Protocol.fds_type_offs);
                nameOffs = Protocol.unpack2(desc, offs + Protocol.fds_name_offs) + Protocol.cds_header_size;
                size = Protocol.unpack4(desc, offs + Protocol.fds_size_offs);
                nItems = Protocol.unpack4(desc, offs + Protocol.fds_nitems_offs);
                next = Protocol.unpack4(desc, offs + Protocol.fds_next_offs);
                fieldName = Protocol.unpackAscii(desc, nameOffs);
				
                offs += Protocol.fds_size;
				
                if (dbsType == Protocol.fld_structure)
                {
                    superclass = new ClassDescriptor(fieldName, desc, offs);
                    n += superclass.nFields;
                    offs += superclass.nFields * Protocol.fds_size;
                    System.Diagnostics.Debug.Assert(firstField == null);
                    varyingField = superclass.varyingField;
                    firstField = superclass.firstField;
                    lastField = superclass.lastField;
                    refFields = superclass.refFields;
                }
                else
                {
                    FieldDescriptor fd = new FieldDescriptor(fieldName, appClass, dbsType, size);
                    if (nItems == 0)
                    {
                        if (!fd.isVarying)
                        {
                            throw new ClassDefinitionError("Field '" + fieldName + "' can not be converted from varying array type");
                        }
                        if (varyingField != null)
                        {
                            throw new ClassDefinitionError("More than one varying components in persistent capable class");
                        }
                        varyingField = fd;
                    }
                    else if (nItems != 1)
                    {
                        throw new ClassDefinitionError("Field '" + fieldName + "' can not be converted from fixed array type");
                    }
                    else if (fd.isVarying)
                    {
                        throw new ClassDefinitionError("Field '" + fieldName + "' can not be converted to varying array type");
                    }
                    if (firstField == null)
                    {
                        firstField = fd;
                    }
                    else
                    {
                        lastField.next = fd;
                    }
                    lastField = fd;
                    if (fd.isVarying || dbsType == Protocol.fld_reference || dbsType == Protocol.fld_string || dbsType == Protocol.fld_raw_binary)
                    {
                        fd.nextRef = refFields;
                        refFields = fd;
                    }
                    if (dbsType == Protocol.fld_string)
                    {
                        fd.nextStr = strFields;
                        strFields = fd;
                    }
                    if (dbsType == Protocol.fld_raw_binary)
                    {
                        fd.nextBin = binFields;
                        binFields = fd;
                    }
                }
                n += 1;
            }
            while (next != 0);
			
            return n;
        }
		
		
        internal ClassDescriptor(String className, byte[] desc, int offs)
        {
            appClass = findClass(className);
            name = className;
            nFields = buildFieldsList(desc, offs);
            setDefaultMetaobject();
        }
				
        internal static ClassDescriptor lookup(Type cls)
        {
            lock(typeof(Goods.ClassDescriptor))
            {
                ClassDescriptor desc = (ClassDescriptor) classDictionary[cls];
                return desc != null ? desc : new ClassDescriptor(cls);
            }
        }
		
        public virtual bool equal(byte[] desc)
        {
            if (desc.Length != dbsClass.Length)
            {
                return false;
            }
            for (int i = desc.Length; --i >= 0; )
            {
                if (desc[i] != dbsClass[i])
                {
                    return false;
                }
            }
            return true;
        }
		
        internal ClassDescriptor(Type cls)
        {
            int i;
            FieldDescriptor fd;
			
            initialize(cls);
            cid = classCount++;
            newClass = this;
			
            System.Reflection.FieldInfo[] classFields = cls.GetFields(System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.DeclaredOnly);
            if (cls.BaseType != typeof(Persistent))
            {
                superclass = lookup(cls.BaseType);
                nFields = 1 + superclass.nFields;
                fixedSize = superclass.fixedSize;
                varyingSize = superclass.varyingSize;
                nFixedReferences = superclass.nFixedReferences;
                nVaryingReferences = superclass.nVaryingReferences;
                totalNamesSize = superclass.totalNamesSize;
				
                for (fd = superclass.firstField; fd != null; fd = fd.next)
                {
                    FieldDescriptor copy = new FieldDescriptor(fd);
                    if (lastField == null)
                    {
                        firstField = copy;
                    }
                    else
                    {
                        lastField.next = copy;
                    }
                    lastField = copy;
                    if (fd == superclass.varyingField)
                    {
                        varyingField = copy;
                    }
                    if (fd.isVarying || fd.dbsType == Protocol.fld_reference || fd.dbsType == Protocol.fld_string || fd.dbsType == Protocol.fld_raw_binary)
                    {
                        copy.nextRef = refFields;
                        refFields = copy;
                    }
                    if (fd.dbsType == Protocol.fld_string)
                    {
                        copy.nextStr = strFields;
                        strFields = copy;
                    }
                    if (fd.dbsType == Protocol.fld_raw_binary)
                    {
                        copy.nextBin = binFields;
                        binFields = copy;
                    }
                }
            }
            totalNamesSize += name.Length + 1;
			
            for (i = 0; i < classFields.Length; i++)
            {
                System.Reflection.FieldInfo f = classFields[i];
                if (f.IsNotSerialized) 
                { 
                    continue;
                }
                fd = new FieldDescriptor(f, cls);
                if (fd.isVarying || fd.dbsType == Protocol.fld_reference || fd.dbsType == Protocol.fld_string || fd.dbsType == Protocol.fld_raw_binary)
                {
                    fd.nextRef = refFields;
                    refFields = fd;
                }
                if (fd.dbsType == Protocol.fld_reference)
                {
                    if (fd.isVarying)
                    {
                        nVaryingReferences = 1;
                    }
                    else
                    {
                        nFixedReferences += 1;
                    }
                }
                else if (fd.dbsType == Protocol.fld_string)
                {
                    fd.nextStr = strFields;
                    strFields = fd;
                }
                else if (fd.dbsType == Protocol.fld_raw_binary)
                {
                    fd.nextBin = binFields;
                    binFields = fd;
                }
                if (fd.isVarying)
                {
                    if (varyingField != null)
                    {
                        throw new ClassDefinitionError("More than one varying components in persistent capable class");
                    }
                    varyingField = fd;
                    varyingSize = fd.size;
                }
                else
                {
                    fixedSize += fd.size;
                }
                if (firstField == null)
                {
                    firstField = fd;
                }
                else
                {
                    lastField.next = fd;
                }
                lastField = fd;
                nFields += 1;
                totalNamesSize += fd.name.Length + 1;
            }
            if (varyingField != null && (strFields != null || binFields != null))
            {
                throw new ClassDefinitionError("Persistent class with String components should not have array component");
            }
            byte[] desc = new byte[totalNamesSize + Protocol.cds_header_size + nFields * Protocol.fds_size];
            Protocol.pack4(desc, Protocol.cds_fixed_size_offs, fixedSize);
            Protocol.pack4(desc, Protocol.cds_varying_size_offs, varyingSize);
            Protocol.pack4(desc, Protocol.cds_n_fixed_refs_offs, nFixedReferences);
            Protocol.pack4(desc, Protocol.cds_n_varying_refs_offs, nVaryingReferences);
            Protocol.pack4(desc, Protocol.cds_n_fields_offs, nFields);
            Protocol.pack4(desc, Protocol.cds_total_names_size_offs, totalNamesSize);
            int nameOffs = Protocol.cds_header_size + nFields * Protocol.fds_size;
            int offs = Protocol.cds_header_size;
            int index = 0;
            int next = nFields;
            nameOffs = Protocol.packAscii(desc, nameOffs, name);
			
            for (ClassDescriptor sup = superclass; sup != null; sup = sup.superclass)
            {
                index += 1;
                Protocol.pack2(desc, offs + Protocol.fds_type_offs, Protocol.fld_structure);
                Protocol.pack2(desc, offs + Protocol.fds_name_offs, nameOffs - Protocol.cds_header_size);
                Protocol.pack4(desc, offs + Protocol.fds_size_offs, sup.fixedSize);
                int succ = index + sup.nFields;
                Protocol.pack4(desc, offs + Protocol.fds_next_offs, succ == next?0:succ);
                Protocol.pack4(desc, offs + Protocol.fds_nitems_offs, 1);
                nameOffs = Protocol.packAscii(desc, nameOffs, sup.name);
                next = succ;
                offs += Protocol.fds_size;
            }
            for (fd = firstField; fd != null; fd = fd.next)
            {
                index += 1;
                Protocol.pack2(desc, offs + Protocol.fds_type_offs, fd.dbsType);
                Protocol.pack2(desc, offs + Protocol.fds_name_offs, nameOffs - Protocol.cds_header_size);
                Protocol.pack4(desc, offs + Protocol.fds_size_offs, fd.size);
                Protocol.pack4(desc, offs + Protocol.fds_nitems_offs, fd == varyingField?0:1);
                Protocol.pack4(desc, offs + Protocol.fds_next_offs, (fd.next == null || fd.next.cls != fd.cls)?0:index);
                nameOffs = Protocol.packAscii(desc, nameOffs, fd.name);
                offs += Protocol.fds_size;
            }
            dbsClass = desc;
            classDictionary[cls] = this;
            conversionNeeded = false;
        }
		
        internal ObjectHeader construct(Storage storage, int opid)
        {
            Persistent p = (Persistent)constructor.Invoke(constructorParameters);
            //    Activator.CreateInstance(constructor.ReflectedType, constructorParameters);
            ObjectHeader obj = p._hdr(); 
            obj.desc = this;
            obj.opid = opid;
            obj.storage = storage;
            obj.state = ObjectState.RAW;
            storage.objectCache.put(opid, obj);
            return obj;
        }
		
		
        internal void  unpack(Persistent obj, ObjectHeader[] refs, int nRefs, int varyingLength)
        {
            FieldDescriptor fd;
            int i, n = 0, offs = Protocol.hdr_size;
            byte[] data = obj.hdr.data;
            obj.hdr.data = null;
			
            if (refs != null)
            {
                offs += nRefs * Protocol.ref_size;
            }
			
            if (!conversionNeeded)
            {
                for (fd = firstField; fd != null; fd = fd.next)
                {
                    if (fd == varyingField)
                    {
                        switch (fd.appType)
                        {
                            case FieldDescriptor.t_sbyte: 
                            {
                                sbyte[] arr = new sbyte[varyingLength];
                                for (i = 0; i < varyingLength; i++) 
                                { 
                                    arr[i] = (sbyte)data[offs++];
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_ubyte: 
                            {
                                byte[] arr = new byte[varyingLength];
                                for (i = 0; i < varyingLength; i++) 
                                { 
                                    arr[i] = data[offs++];
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }

                            case FieldDescriptor.t_short: 
                            {
                                short[] arr = new short[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (short) Protocol.unpack2(data, offs);
                                    offs += 2;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_ushort: 
                            {
                                ushort[] arr = new ushort[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (ushort) Protocol.unpack2(data, offs);
                                    offs += 2;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_integer: 
                            {
                                int[] arr = new int[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = Protocol.unpack4(data, offs);
                                    offs += 4;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_uinteger: 
                            {
                                uint[] arr = new uint[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (uint)Protocol.unpack4(data, offs);
                                    offs += 4;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_ulong: 
                            {
                                ulong[] arr = new ulong[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (ulong)Protocol.unpack8(data, offs);
                                    offs += 8;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_long: 
                            {
                                long[] arr = new long[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = Protocol.unpack8(data, offs);
                                    offs += 8;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_float: 
                            {
                                float[] arr = new float[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = BitConverter.ToSingle(BitConverter.GetBytes(Protocol.unpack4(data, offs)), 0);                                       
                                    offs += 4;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_double: 
                            {
                                double[] arr = new double[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = BitConverter.Int64BitsToDouble(Protocol.unpack8(data, offs));                                            
                                    offs += 8;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_char: 
                            {
                                char[] arr = new char[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = (char) Protocol.unpack2(data, offs);
                                    offs += 2;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_boolean: 
                            {
                                bool[] arr = new bool[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    arr[i] = data[offs++] != 0;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_reference: 
                            {
                                Array arr = Array.CreateInstance(fd.field.FieldType.GetElementType(), varyingLength);
                                for (i = 0; i < varyingLength; i++) 
                                {
                                    ObjectHeader r = refs[i+n];
                                    if (r != null) 
                                    { 
                                        arr.SetValue(r.obj, i);
                                    }
                                }
                                n += varyingLength;
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                            case FieldDescriptor.t_date: 
                            {
                                DateTime[] arr = new DateTime[varyingLength];
                                for (i = 0; i < varyingLength; i++)
                                {
                                    long val = Protocol.unpack8(data, offs);
                                    arr[i] = Protocol.NewDateTime(val);
                                    offs += 8;
                                }
                                fd.field.SetValue(obj, arr);
                                break;
                            }
								
                        }
                    }
                    else
                    {
                        switch (fd.appType)
                        {
                            case FieldDescriptor.t_sbyte: 
                                fd.field.SetValue(obj, (sbyte)data[offs++]);
                                break;
								
                            case FieldDescriptor.t_ubyte: 
                                fd.field.SetValue(obj, data[offs++]);
                                break;
								
                            case FieldDescriptor.t_integer: 
                                fd.field.SetValue(obj, Protocol.unpack4(data, offs));
                                offs += 4;
                                break;
								
                            case FieldDescriptor.t_uinteger: 
                                fd.field.SetValue(obj, (uint)Protocol.unpack4(data, offs));
                                offs += 4;
                                break;
								
                            case FieldDescriptor.t_long: 
                                fd.field.SetValue(obj, Protocol.unpack8(data, offs));
                                offs += 8;
                                break;
								
                            case FieldDescriptor.t_ulong: 
                                fd.field.SetValue(obj, (ulong)Protocol.unpack8(data, offs));
                                offs += 8;
                                break;
								
                            case FieldDescriptor.t_short: 
                                fd.field.SetValue(obj, (short)Protocol.unpack2(data, offs));
                                offs += 2;
                                break;
								
                            case FieldDescriptor.t_ushort: 
                                fd.field.SetValue(obj, (ushort)Protocol.unpack2(data, offs));
                                offs += 2;
                                break;
								
                            case FieldDescriptor.t_char: 
                                fd.field.SetValue(obj, (char) Protocol.unpack2(data, offs));
                                offs += 2;
                                break;
								
                            case FieldDescriptor.t_float: 
                                fd.field.SetValue(obj, BitConverter.ToSingle(BitConverter.GetBytes(Protocol.unpack4(data, offs)), 0));   
                                offs += 4;
                                break;
								
                            case FieldDescriptor.t_double: 
                                fd.field.SetValue(obj, BitConverter.Int64BitsToDouble(Protocol.unpack8(data, offs)));                                            
                                offs += 8;
                                break;
								
                            case FieldDescriptor.t_boolean: 
                                fd.field.SetValue(obj, data[offs++] != 0);
                                break;
								
                            case FieldDescriptor.t_reference: 
                                ObjectHeader r = refs[n++];
                                fd.field.SetValue(obj, r != null ? r.obj : null);
                                break;
								
                            case FieldDescriptor.t_string: 
                                i = Protocol.unpack2(data, offs) & 0xFFFF;
                                offs += 2;
                                if (i == FieldDescriptor.nullString)
                                {
                                    fd.field.SetValue(obj, null);
                                }
                                else
                                {
                                    char[] arr = new char[i];
                                    for (int j = 0; j < i; j++)
                                    {
                                        arr[j] = (char) Protocol.unpack2(data, offs);
                                        offs += 2;
                                    }
                                    String s = new String(arr);
                                    fd.field.SetValue(obj, s);
                                }
                                break;
								
                            case FieldDescriptor.t_raw_binary: 
                                i = Protocol.unpack4(data, offs);
                                offs += 4;
                                fd.field.SetValue(obj, Database.rawBinaryFactory.unpack(fd.field.FieldType, data, offs, i));
                                offs += i;
                                break;
								
                            case FieldDescriptor.t_date: 
                            {
                                long val = Protocol.unpack8(data, offs);
                                fd.field.SetValue(obj, Protocol.NewDateTime(val));
                                offs += 8;
                                break;
                            }								
                        }
                    }
                }
            }
            else
            {
                for (fd = firstField; fd != null; fd = fd.next)
                {
                    if (fd == varyingField)
                    {
                        if (fd.dbsType == Protocol.fld_reference)
                        {
                            if (fd.field != null)
                            {
                                Array arr = Array.CreateInstance(fd.field.FieldType.GetElementType(), varyingLength);
                                for (i = 0; i < varyingLength; i++) 
                                { 
                                    ObjectHeader r = refs[i + n];
                                    if (r != null) 
                                    { 
                                        arr.SetValue(r.obj, i);
                                    }
                                }
                                fd.field.SetValue(obj, arr);
                            }
                            n += varyingLength;
                        }
                        else
                        {
                            offs = varyingField.unpackVarying(data, offs, obj, varyingLength);
                        }
                    }
                    else
                    {
                        if (fd.dbsType == Protocol.fld_reference)
                        {
                            if (fd.field != null)
                            {
                                ObjectHeader r = refs[n];
                                fd.field.SetValue(obj, r != null ? r.obj : null);
                            }
                            n += 1;
                        }
                        else
                        {
                            offs = fd.unpack(data, offs, obj);
                        }
                    }
                }
            }
            obj.hdr.state &= ~ObjectState.RAW;
        }
		
		
        internal int objectSize(Persistent obj)
        {
            int size = fixedSize;
            if (varyingField != null)
            {
                size += varyingField.length(obj) * varyingSize;
            }
            for (FieldDescriptor str = strFields; str != null; str = str.nextStr)
            {
                if (str.field != null)
                {
                    String s = (String)str.field.GetValue(obj);
                    if (s != null)
                    {
                        size += s.Length * 2;
                    }
                }
            }
            for (FieldDescriptor bin = binFields; bin != null; bin = bin.nextBin)
            {
                if (bin.field != null)
                {
                    Object rbo = bin.field.GetValue(obj);
                    if (rbo != null)
                    {
                        size += Database.rawBinaryFactory.size(rbo);
                    }
                }
            }
            return size;
        }
		
        internal void  pack(Persistent obj, byte[] buffer, int offs)
        {
            int varyingLength = (varyingField != null) ? varyingField.length(obj) : 0;
            int scalarOffs = offs + Protocol.ref_size * (nFixedReferences + varyingLength * nVaryingReferences);
			
            for (FieldDescriptor fd = firstField; fd != null; fd = fd.next)
            {
                if (fd == varyingField)
                {
                    if (fd.dbsType == Protocol.fld_reference)
                    {
                        offs = varyingField.packObjectArray(obj, buffer, offs);
                    }
                    else
                    {
                        scalarOffs = varyingField.packVarying(obj, buffer, scalarOffs);
                    }
                }
                else
                {
                    if (fd.dbsType == Protocol.fld_reference)
                    {
                        offs = fd.packRef(obj, buffer, offs);
                    }
                    else
                    {
                        scalarOffs = fd.pack(obj, buffer, scalarOffs);
                    }
                }
            }
        }

        internal void reset(Persistent obj) 
        {
            for (FieldDescriptor fd = refFields; fd != null; fd = fd.nextRef)
            {
                fd.field.SetValue(obj, null);
            }   
        }
    }
}