//-< PessimisticRepeatableReadMetaobject.java >----------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  9-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Metaobject for pessimistic scheme of synchronization with repeatable read
// isolation level
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary>  <code>PessimisticRepeatableReadMetaobject</code>  for pessimistic scheme of
    /// synchronization with repeatable read  isolation level 
    /// While extending the Multiple Reader, one writer aproach, this strategy obtains
    /// shred locks for reading, thus ensuring that no write is going on at the same time.
    /// </summary>
    /// <author>  K.A. Knizhnik 
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public class PessimisticRepeatableReadMetaobject:PessimisticMetaobject
    {
        /// <summary> <code>beginReadAccess</code> locks the object in shared mode.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to be read.
        /// 
        /// </param>
        protected internal override void  beginReadAccess(ObjectHeader obj)
        {
            if ((obj.state & (ObjectState.SLOCKED | ObjectState.XLOCKED)) == 0)
            {
                obj.storage.lockObject(obj, Protocol.lck_shared, Protocol.lckattr_wait);
                obj.state |= ObjectState.SLOCKED;
            }
        }
		
        /// <summary>  <code>endAccess</code> adds the object to the transaction list, also when the 
        /// object was not modified. Thus the (shared) lock is released.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object 
        /// 
        /// </param>
        protected internal override void  endAccess(ObjectHeader obj)
        {
            ObjectState state = obj.state;
            if ((state & ObjectState.TRANSWRITE) == 0 && (state & (ObjectState.DIRTY | ObjectState.XLOCKED)) != 0)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                mng.addToTransaction(obj);
            }
            else if ((state & (ObjectState.TRANSREAD | ObjectState.TRANSWRITE | ObjectState.SLOCKED)) == ObjectState.SLOCKED)
            {
                System.Diagnostics.Debug.Assert((state & ObjectState.DIRTY) == 0);
                CacheManager mng = CacheManager.GetCacheManager();
                mng.addToTransaction(obj);
            }
        }
    }
}