//-< Protocol.java >-------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 10-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Description of client-server communication protocol
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    class Protocol
    {
        internal const int RAW_CPID = 0x00001;
        internal const int ROOT_OPID = 0x10000;
		
        internal const int ALLOC_ALIGNED = (int) (- (0x100000000 - 0x80000000));
		
        //
        // Object reference components
        //
        internal const int ref_sid_offs = 0;
        internal const int ref_opid_offs = 2;
		
        internal const int ref_size = 6;
		
        //
        // Offsets of object header fields
        //
        internal const int hdr_flags_offs = 0;
        internal const int hdr_opid_offs = 4;
        internal const int hdr_cpid_offs = 8;
        internal const int hdr_sid_offs = 10;
        internal const int hdr_size_offs = 12;
		
        internal const int hdr_size = 16;
		
        //
        // Types of object fields
        //
        internal const int fld_structure = 0;
        internal const int fld_reference = 1;
        internal const int fld_signed_integer = 2;
        internal const int fld_unsigned_integer = 3;
        internal const int fld_real = 4;
        internal const int fld_string = 5;
        internal const int fld_raw_binary = 6;
		
        //
        // Field descriptor fields offsets
        //
        internal const int fds_type_offs = 0;
        internal const int fds_name_offs = 2;
        internal const int fds_size_offs = 4;
        internal const int fds_nitems_offs = 8;
        internal const int fds_next_offs = 12;
		
        internal const int fds_size = 16;
		
        //
        // Class descriptor fields offsets
        //
        internal const int cds_fixed_size_offs = 0;
        internal const int cds_varying_size_offs = 4;
        internal const int cds_n_fixed_refs_offs = 8;
        internal const int cds_n_varying_refs_offs = 12;
        internal const int cds_n_fields_offs = 16;
        internal const int cds_total_names_size_offs = 20;
		
        internal const int cds_header_size = 24;
		
        //
        // Lock request types
        //
        internal const int lck_none = 0; // No lock
        internal const int lck_shared = 1; // Prevent object from been modified or
        // exclusivly locked by other processes
        internal const int lck_exclusive = 2; // Prevent object from been modified or
        // locked (shared or exlusive) by other processes
		
        //
        // Lock attributes
        //
        internal const int lckattr_wait = 0;
        internal const int lckattr_nowait = 1; // Lock request will not block, 
        // if granting lock is impossible, request will be immediately refused 
		
        //
        // Transaction object flags 
        //
        internal const int tof_none = 0;
        internal const int tof_update = 1; // request to update object
        internal const int tof_validate = 2; // request to validate transaction object at server
        internal const int tof_unlock = 4; // request to unlock transaction object at the end of transaction
        internal const int tof_unlock_exl = 8; // request to release exclusive lock at the end of transaction
        //
        // load object flags
        //
        internal const int lof_none = 0;
        internal const int lof_copy = 1; // request another copy of object
        internal const int lof_cluster = 2; // enable sending cluster of objects
		
		
        // 
        // Command request codes
        // 
        internal const byte cmd_load = 0; // client request object from server
        internal const byte cmd_object = 1; // server sent object to client
        internal const byte cmd_forget = 2; // client remove reference to object
        internal const byte cmd_throw = 3; // client throw away instance of object from the cache
        internal const byte cmd_invalidate = 4; // server inform client about object modification
		
        internal const byte cmd_getclass = 5; // get class descriptor from server
        internal const byte cmd_classdesc = 6; // server send class descriptor to client
		
        internal const byte cmd_putclass = 7; // register new class at server 
        internal const byte cmd_classid = 8; // class identifier returned to client 
		
        internal const byte cmd_modclass = 9; // modify existed class at server
		
        internal const byte cmd_lock = 10; // request to server to lock object
        internal const byte cmd_lockresult = 11; // result of applying lock request 
        internal const byte cmd_unlock = 12; // request to server to unlock object
		
        internal const byte cmd_transaction = 13; // send transaction to coordinator
        internal const byte cmd_subtransact = 14; // send local part of global trans.
        internal const byte cmd_transresult = 15; // coordinator returns trans. status
        internal const byte cmd_tmsync = 16; // byteerserver transaction sync. message
		
        internal const byte cmd_login = 17; // client login at server 
        internal const byte cmd_logout = 18; // client finish the session
        internal const byte cmd_connect = 19; // one server connects to other 
        internal const byte cmd_bye = 20; // server notifies client  about it's termination
        internal const byte cmd_ok = 21; // successful authorization 
        internal const byte cmd_refused = 22; // authorization procedure is failed at server
        internal const byte cmd_alloc = 23; // request server to allocate object
        internal const byte cmd_location = 24; // server returns address of allocated object
        internal const byte cmd_free = 25; // client request to free object
        internal const byte cmd_bulk_alloc = 26; // client request server to allocate several objects and reserve OIDs 
        internal const byte cmd_opids = 27; // object identifiers reserved by server for the client
        internal const byte cmd_gcsync = 28; // interserver garbage collection synchonization
		
        internal const byte cmd_get_size = 29; // get storage size
		
        //
        // Request fields offsets
        //
        internal const int req_cmd_offs = 0;
        internal const int req_arg1_offs = 1;
        internal const int req_arg2_offs = 2;
        internal const int req_arg3_offs = 4;
        internal const int req_arg4_offs = 8;
        internal const int req_arg5_offs = 12;
		
        internal const int req_type_offs = 1;
        internal const int req_n_servers_offs = 1;
        internal const int req_align_offs = 1;
        internal const int req_status_offs = 1;
        internal const int req_coordinator_offs = 2;
        internal const int req_flags_offs = 2;
        internal const int req_cpid_offs = 2;
        internal const int req_sid_offs = 2;
        internal const int req_attr_offs = 2;
        internal const int req_size_offs = 4;
        internal const int req_opid_offs = 4;
        internal const int req_extra_offs = 8;
        internal const int req_reserve_offs = 8;
        internal const int req_tid_offs = 8;
		
        internal const int req_size = 16;
		
        internal static int unpack2(byte[] data, int offs)
        {
            return ((sbyte)data[offs] << 8) + data[offs + 1];
        }
        internal static int unpack4(byte[] data, int offs)
        {
            return ((sbyte)data[offs] << 24) + (data[offs + 1] << 16) + (data[offs + 2] << 8) + data[offs + 3];
        }
        internal static long unpack8(byte[] data, int offs)
        {
            return ((long)unpack4(data, offs) << 32) + (uint)unpack4(data, offs + 4);
        }
        internal static String unpackAscii(byte[] data, int offs)
        {
            int eos = offs;
            while (data[eos] != 0)
            {
                eos += 1;
            }
            char[] chars = new char[eos-offs];
            for (int i = 0; i < chars.Length; i++) 
            { 
                chars[i] = (char)data[offs+i];
            }
            return new String(chars);
        }
		
        internal static int pack2(byte[] data, int offs, int val)
        {
            data[offs++] = (byte) (val >> 8);
            data[offs++] = (byte) val;
            return offs;
        }
        internal static int pack4(byte[] data, int offs, int val)
        {
            data[offs++] = (byte) (val >> 24);
            data[offs++] = (byte) (val >> 16);
            data[offs++] = (byte) (val >> 8);
            data[offs++] = (byte) val;
            return offs;
        }
        internal static int pack8(byte[] data, int offs, long val)
        {
            return pack4(data, pack4(data, offs, (int) (val >> 32)), (int) val);
        }
		
        internal static int packAscii(byte[] data, int offs, String s)
        {
            int length = s.Length;
            int i;
            int j;
            i = 0;
            j = offs;
            while (i < length)
            {
                data[j] = (byte) s[i];
                i++;
                j++;
            }
            data[offs + length] = 0;
            return offs + length + 1;
        }


        internal static int packDate(byte[] data, int offs, DateTime dt) 
        { 
            return pack8(data, offs, dt.Ticks == -1 ? dt.Ticks : (dt.Ticks - 621355968000000000L) / 10000);
        }

        internal static DateTime NewDateTime(long milliseconds) 
        {
            return new DateTime(milliseconds == -1 ? milliseconds : milliseconds*10000 + 621355968000000000L);
        }         
    }
}