//-< Storage.java >--------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  4-Nov-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Client interface with storage server
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
    using System.Net.Sockets;
    using System.Net;
	
    /// <summary> Client interface with storage server
    /// </summary>
    public class Storage 
    {
        protected internal virtual Persistent Root
        {
            //
            // Retrieve reference to root object in the storage. Each GOODS
            // storage has preexested abstract root object. To prevent concurrent
            // initialization of storage by several clients, object with root PID
            // is locked and if this object is abstract root, null is returned to
            // client application allowing it to astablish it's own root object
            // as storage root (lcok prevent other clients from doing the same).
            // If root object is not abstract root, then it is unlocked and reference
            // to it is returnednto the client.
            // 
            get
            {
                lock(this)
                {
                    if (root == null)
                    {
                        root = new ObjectHeader(null);
                        root.opid = Protocol.ROOT_OPID;
                        lockObject(root, Protocol.lck_exclusive, Protocol.lckattr_wait);
                        root.opid = 0;
                        root = load(Protocol.ROOT_OPID, null);
                        if (root != null)
                        {
                            unlock(root, Protocol.lck_none);
                        }
                    }
                    return root != null ? root.obj : null;
                }
            }
			
            //
            // Set root for the storage. This method can only be called when previously
            // invoked getRoot() method returns null. It means that storage root 
            // can be set only once.
            //
            set
            {
                lock(this)
                {
                    ObjectHeader hdr = value._hdr();
                    System.Diagnostics.Debug.Assert(hdr.opid == 0, "only transient object can become a root");
                    System.Diagnostics.Debug.Assert(hdr.accessCount == 0, "class was preprocessed by JavaMOP");
                    ObjectHeader oldRoot = (ObjectHeader) objectCache.get(Protocol.ROOT_OPID);
                    System.Diagnostics.Debug.Assert(oldRoot == null, "root object can't be changed");
					
                    hdr.opid = Protocol.ROOT_OPID;
                    hdr.storage = this;
                    hdr.state |= ObjectState.DIRTY | ObjectState.XLOCKED;
                    objectCache.put(Protocol.ROOT_OPID, hdr);
                    hdr.metaobject.preDaemon(hdr, Metaobject.MUTATOR);
                    hdr.metaobject.postDaemon(hdr, Metaobject.MUTATOR);
                    root = hdr;
                }
            }			
        }
        protected internal virtual long Size
        {
            get
            {
                lock(this)
                {
                    request[Protocol.req_cmd_offs] = Protocol.cmd_get_size;
                    sendReceive(request, Protocol.cmd_get_size);
					
                    int high = Protocol.unpack4(result, Protocol.req_arg3_offs);
                    int low = Protocol.unpack4(result, Protocol.req_arg4_offs);
					
                    return (low & (long) 0xffffffff) + ((long) high << 32);
                }
            }
			
        }
        protected internal int id;
        protected internal Database database;
		
        protected internal Socket socket;
        public static int LingerTime = 10; // linger paramter for the socket
		
        protected internal byte[] result; // buffer for passing server reply to the client
        protected internal byte[] notification; // buffer for sending notifications to server
        protected internal byte[] reply; // buffer for soring reply received from server
        protected internal byte[] request; // buffer for sending single request to server
		
        //
        // Array to establish mapping between  persistent class indentifiers 
        // and application class descriptors
        //
        internal ClassDescriptor[] classDictionary;
        protected internal int classDictionarySize;
		
        protected internal int[] cpidTable;
        protected internal int cpidTableSize;

        internal class AllocBuf
        {
            internal int opid;
            internal int cpid;
            internal int size;
        }
        internal AllocBuf[] bulkAllocBuf;
        internal int allocBufPos;
        internal int allocBufSize;
		
        // 
        // The following variables are used to keep list of stubs for just loaded
        // objects. These stubs solve two problems: 
        // 1) avoiding load of the same object more than once
        // 2) handling invalidation message for the object during loading
        //
        protected internal ObjectHeader loadList;
        protected internal ObjectHeader freeList;
		
        protected internal Object cs; // Critiacal section for socket write operations
        protected internal Object requestMutex; // Critiacal section for send-receive operations
    		
        protected internal ObjectHeader root; // Root object in the storage
		
        //
        // These event objects are used to synchronize receiver thread with 
        // client application threads. Receiver will handle asynchronouse 
        // notifications from server itself and signal rcvEvent object when reply 
        // for client request arriives. Then receiver waits until client signal
        // repEvent object to continue receving loop.
        //
        protected internal CondEvent rcvEvent;
        protected internal CondEvent repEvent;
		
        //
        // Buffer for storing body of transaction. If size of transaction buffer 
        // exceeds maxTransBufferSize, then reference to it will be cleared
        // after the end oftransaction to allow GC to free space occupid by this
        // buffer. Otherwise reference to buffer is kept to reuse it in following
        // transactions.
        //
        protected internal byte[] transBuffer;
        protected internal int transBufferOffs;
		
        public static int initTransBufferSize = 64 * 1024;
        public static int maxTransBufferSize = 1024 * 1024;
		
        //
        // Weak cache of objects. This cache is used to locate loaded instances of 
        // persistent objects. Combination of object storage identifier and object
        // persistent identifier is used as key value for putting/searching
        // objects in cache. Weakness of the cache means that GC is able to
        // free object referenced only from this cache class.
        //
        protected internal ObjectCache objectCache;
		
		
        internal const int MAX_REFS = 256;
        internal byte[] smallReqbuf;
        internal int[] smallPermutation;
        internal int[] smallSids;
        internal ObjectHeader[] smallRefs;
		
        protected internal virtual void  createWeakObjectCache()
        {
            objectCache = new WeakHashTable();
        }
		
        //
        // inform receiver about close initiation
        //
        protected internal volatile bool closing;
		
        protected internal System.Threading.Thread receiver;
		
        //
        // Check cache and return object, or go and load it from db
        //
        protected internal virtual Persistent getObject(int id)
        {
            ObjectHeader hdr = (ObjectHeader) objectCache.get(id);
            if (hdr == null)
            {
                hdr = load(id, null);
            }
            return hdr != null ? hdr.obj : null;
        }
		
        //
        // Associate persistent class identifier with application class indentifier
        //
        internal void  associateCpidWithClass(ClassDescriptor desc, int cpid)
        {
            if (desc.cid >= cpidTableSize)
            {
                int newSize = desc.cid >= cpidTableSize * 2?desc.cid + 1:cpidTableSize * 2;
                int[] newTable = new int[newSize];
                Array.Copy(cpidTable, 0, newTable, 0, cpidTableSize);
                cpidTable = newTable;
                cpidTableSize = newSize;
            }
            cpidTable[desc.cid] = cpid;
        }
		
        // 
        // Extend size of transaction buffer 
        //
        internal void  appendTransBuffer(int size)
        {
            if (transBufferOffs + size > transBuffer.Length)
            {
                int newSize = transBufferOffs + size > transBuffer.Length * 2?transBufferOffs + size:transBuffer.Length * 2;
                byte[] newBuffer = new byte[newSize];
                Array.Copy(transBuffer, 0, newBuffer, 0, transBufferOffs);
                transBuffer = newBuffer;
            }
            transBufferOffs += size;
        }
		
        //
        // Load object instances and initialize variables for detecting
        // deterioration of just loaded objects
        //
        internal byte[] loadObjects(byte[] buf, int nObjects)
        {
            Protocol.pack4(buf, Protocol.req_extra_offs, nObjects - 1);
            return sendReceiveBody(buf, nObjects * Protocol.req_size, Protocol.cmd_object);
        }
		
        //
        // Load all objects referenced from specified object "obj". 
        // When object is loaded from database all its references are replaced 
        // with stub objects which type corresponds to the static type of field.
        // That stub object are used only for storing OPIDs of referenced objects.
        // Object is marked as RAW. Whan it is then accessed by its method 
        // invokation or component store/fetch operation, prepare() method
        // is called by metaobject to replace stub objects with real objects.
        //
        internal void  prepare(ObjectHeader obj)
        {
            lock(this)
            {
                ClassDescriptor desc = obj.desc;
                byte[] data = obj.data;
                int i, j, k;
				
                System.Diagnostics.Debug.Assert((obj.state & (ObjectState.DESTRUCTED | ObjectState.RAW)) == ObjectState.RAW && data != null && data.Length > 0);
				
                int size = Protocol.unpack4(data, Protocol.hdr_size_offs);
                int varyingLength = 0;
                if (desc.varyingSize != 0)
                {
                    varyingLength = (size - desc.fixedSize) / desc.varyingSize;
                }
                int nRefs = desc.nFixedReferences + varyingLength * desc.nVaryingReferences;
                ObjectHeader[] refs = null;
                if (nRefs > 0)
                {
                    byte[] reqbuf;
                    int[] permutation;
                    int[] sids;
					
                    if (nRefs <= MAX_REFS)
                    {
                        refs = smallRefs;
                        reqbuf = smallReqbuf;
                        permutation = smallPermutation;
                        sids = smallSids;
                    }
                    else
                    {
                        refs = new ObjectHeader[nRefs];
                        reqbuf = new byte[nRefs * Protocol.req_size];
                        permutation = new int[nRefs];
                        sids = new int[nRefs];
                    }
					
                    int offs = Protocol.hdr_size;
                    for (i = 0; i < nRefs; i++)
                    {
                        sids[i] = Protocol.unpack2(data, offs + Protocol.ref_sid_offs);
                        offs += Protocol.ref_size;
                    }
                    for (i = database.storages.Length; --i >= 0; )
                    {
                        offs = Protocol.hdr_size;
                        Storage storage = database.storages[i];
                        lock(storage)
                        {
                            ObjectHeader head = storage.loadList;
                            ObjectHeader free = storage.freeList;
                            head.next = head.prev = head;
                            for (j = 0, k = 0; j < nRefs; j++)
                            {
                                if (sids[j] == i)
                                {
                                    int opid = Protocol.unpack4(data, offs + Protocol.ref_opid_offs);
                                    if (opid != 0)
                                    {
                                        if ((refs[j] = (ObjectHeader) storage.objectCache.get(opid)) == null)
                                        {
                                            int pos = k * Protocol.req_size;
                                            reqbuf[pos + Protocol.req_cmd_offs] = Protocol.cmd_load;
                                            Protocol.pack2(reqbuf, pos + Protocol.req_flags_offs, 0);
                                            Protocol.pack4(reqbuf, pos + Protocol.req_opid_offs, opid);
                                            permutation[k++] = j;
                                            ObjectHeader stub = free;
                                            if (stub == null)
                                            {
                                                stub = new ObjectHeader(null);
                                            }
                                            else
                                            {
                                                free = free.next;
                                            }
                                            //
                                            // Remember object index in "refs" array to
                                            // make it possible to resolve other 
                                            // references to this object at the end of 
                                            // loading. 
                                            //
                                            stub.opid = j;
                                            stub.invalidated = false;
                                            stub.next = head;
                                            stub.prev = head.prev;
                                            head.prev = stub;
                                            stub.prev.next = stub;
                                            storage.objectCache.put(opid, stub);
                                        }
                                    }
                                }
                                offs += Protocol.ref_size;
                            }
                            if (k > 0)
                            {
                                ObjectHeader p;
                                byte[] body = storage.loadObjects(reqbuf, k);
                                head = head.next;
                                offs = 0;
                                for (j = 0; j < k; j++)
                                {
                                    p = storage.readObject(null, body, offs);
                                    refs[permutation[j]] = p;
                                    ObjectHeader stub = head;
                                    head = head.next;
                                    if (stub.invalidated)
                                    {
                                        p.state |= ObjectState.DESTRUCTED;
                                        storage.throwObject(p);
                                        stub.invalidated = false;
                                    }
                                    stub.next = free;
                                    free = stub;
                                    offs += Protocol.hdr_size + Protocol.unpack4(body, offs + Protocol.hdr_size_offs);
                                }
                                for (j = 0; j < nRefs; j++)
                                {
                                    p = refs[j];
                                    if (p != null && p.storage == null)
                                    {
                                        // stub
                                        //
                                        // Stub object contains in "opid" field index
                                        // of real reference to the object in "refs"
                                        // array 
                                        //
                                        refs[j] = refs[p.opid];
                                    }
                                }
                                storage.freeList = free;
                            }
                        }
                    }
                }
                obj.metaobject.loadObject(obj);
                obj.unpack(refs, nRefs, varyingLength);
                for (i = 0; i < nRefs; i++)
                {
                    refs[i] = null;
                }
                obj.desc = desc.newClass;
            }
        }
		
        // 
        // Load object instance from the server. After this method object
        // is in RAW state and prepare() should be called to resolve its references
        //
        protected internal virtual ObjectHeader load(int opid, ObjectHeader obj)
        {
            lock(this)
            {
                byte[] req = new byte[Protocol.req_size];
                req[Protocol.req_cmd_offs] = Protocol.cmd_load;
                Protocol.pack4(req, Protocol.req_opid_offs, opid);
                Protocol.pack2(req, Protocol.req_flags_offs, 0);
                byte[] body = loadObjects(req, 1);
                ObjectHeader p = readObject(obj, body, 0);
                if (p != null)
                {
                    System.Diagnostics.Debug.Assert(obj == p || obj == null);
                    if (req[Protocol.req_cmd_offs] == Protocol.cmd_invalidate)
                    {
                        p.invalidated = true;
                    }
                    prepare(p);
                }
                return p;
            }
        }
		
        //
        // Append object or just object header(for unmodified objects) to 
        // the transaction buffer
        //
        protected internal virtual void  store(ObjectHeader obj)
        {
            ClassDescriptor desc = obj.desc;
            System.Diagnostics.Debug.Assert(obj.opid != 0 && obj.storage == this);
            int cpid = 0;
            if (desc.cid < cpidTableSize)
            {
                cpid = cpidTable[desc.cid];
            }
            if (cpid == 0)
            {
                cpid = registerClass(desc.dbsClass);
                System.Diagnostics.Debug.Assert(cpid != 0);
                associateCpidWithClass(desc, cpid);
            }
            if (transBufferOffs == 0)
            {
                if (transBuffer == null)
                {
                    transBuffer = new byte[initTransBufferSize];
                }
                transBufferOffs = Protocol.req_size;
            }
            int offs = transBufferOffs;
            int flags = 0;
            int size = 0;
			
            if ((obj.state & ObjectState.DIRTY) != 0)
            {
                size = obj.size();
                flags = Protocol.tof_update;
                appendTransBuffer(Protocol.hdr_size + size);
                obj.pack(transBuffer, offs + Protocol.hdr_size);
            }
            else
            {
                appendTransBuffer(Protocol.hdr_size);
            }
            if ((obj.state & ObjectState.OPTIMISTIC) != 0)
            {
                flags |= Protocol.tof_validate;
            }
            if ((obj.state & (ObjectState.XLOCKED | ObjectState.SLOCKED)) != 0)
            {
                flags |= Protocol.tof_unlock;
            }
            Protocol.pack4(transBuffer, offs + Protocol.hdr_flags_offs, flags);
            Protocol.pack4(transBuffer, offs + Protocol.hdr_opid_offs, obj.opid);
            Protocol.pack2(transBuffer, offs + Protocol.hdr_sid_offs, id);
            Protocol.pack2(transBuffer, offs + Protocol.hdr_cpid_offs, cpid);
            Protocol.pack4(transBuffer, offs + Protocol.hdr_size_offs, size);
        }
		
        //
        // Restore original state of transactio buffer. No request should be sent
        // to the server beacuse it knows nothing about this transaction.
        //
        protected internal virtual void  abortTransaction()
        {
            if (transBuffer != null && transBuffer.Length > maxTransBufferSize)
            {
                // Buffer is too large, it is better to let GC to resue it space
                transBuffer = null;
            }
            transBufferOffs = 0;
        }
		
        protected internal virtual bool isInvolvedInTransaction()
        {
            return transBufferOffs != 0;
        }
		
        //
        // Establish mapping between persistent and application classes for
        // loaded object and ask ClassDescriptor to unpack the object
        //
        protected internal virtual ObjectHeader readObject(ObjectHeader obj, byte[] buffer, int offs)
        {
            int cpid = Protocol.unpack2(buffer, offs + Protocol.hdr_cpid_offs);
            int opid = Protocol.unpack4(buffer, offs + Protocol.hdr_opid_offs);
			
            if (cpid == 0)
            {
                throw new ApplicationException("Object with opid " + opid + " not found at server");
            }
			
            if (cpid == Protocol.RAW_CPID)
            {
                return null;
            }
			
            if (cpid >= classDictionarySize)
            {
                int newClassDictionarySize = (classDictionarySize * 2 > cpid)?classDictionarySize * 2:cpid + 1;
                ClassDescriptor[] newClassDictionary = new ClassDescriptor[newClassDictionarySize];
                Array.Copy(classDictionary, 0, newClassDictionary, 0, classDictionarySize);
                classDictionary = newClassDictionary;
                classDictionarySize = newClassDictionarySize;
            }
            ClassDescriptor desc = classDictionary[cpid];
			
            if (desc == null)
            {
                byte[] dbsClassDesc = getClass(cpid);
                Type cls = ClassDescriptor.getClass(dbsClassDesc);
                desc = ClassDescriptor.lookup(cls);
                if (!desc.equal(dbsClassDesc))
                {
                    associateCpidWithClass(desc, registerClass(desc.dbsClass));
                    desc = new ClassDescriptor(desc, dbsClassDesc);
                }
                else
                {
                    associateCpidWithClass(desc, cpid);
                }
                classDictionary[cpid] = desc;
            }
            if (obj == null)
            {
                obj = desc.construct(this, opid);
            }
            else
            {
                obj.desc = desc;
                obj.state = (obj.state & ~ ObjectState.DESTRUCTED) | ObjectState.RAW;
            }
            int size = Protocol.unpack4(buffer, offs + Protocol.hdr_size_offs);
            obj.metaobject.preloadObject(obj, size + Protocol.hdr_size);
            if (size + Protocol.hdr_size != buffer.Length)
            {
                obj.data = new byte[size + Protocol.hdr_size];
                Array.Copy(buffer, offs, obj.data, 0, size + Protocol.hdr_size);
            }
            else
            {
                obj.data = buffer;
            }
            return obj;
        }
		
		
        internal void  invalidate(int opid)
        {
            ObjectHeader obj = (ObjectHeader) objectCache.get(opid);
            if (obj != null)
            {
                Metaobject metaobject = obj.metaobject;
                if (metaobject != null)
                {
                    database.cacheManager.attach(receiver);
                    metaobject.invalidateObject(obj);
                }
                else
                {
                    obj.invalidated = true; // stub 
                }
            }
        }
		
        //
        // Receiver thread
        //
        public void  Run()
        {
            while (true)
            {
                read(reply, Protocol.req_size);
                switch (reply[Protocol.req_cmd_offs])
                {
                    case Protocol.cmd_invalidate: 
                        invalidate(Protocol.unpack4(reply, Protocol.req_opid_offs));
                        int n = Protocol.unpack4(reply, Protocol.req_extra_offs);
                        if (n != 0)
                        {
                            byte[] buf = new byte[n * Protocol.req_size];
                            read(buf, n * Protocol.req_size);
                            int offs = 0;
                            while (--n >= 0)
                            {
                                invalidate(Protocol.unpack4(buf, offs + Protocol.req_opid_offs));
                                offs += Protocol.req_size;
                            }
                        }
                        break;
					
                    case Protocol.cmd_bye: 
                        if (!closing)
                        {
                            database.disconnected(id);
                        }
                        repEvent.signal();
                        return ;
					
                    default: 
                        repEvent.signal();
                        rcvEvent.waitSignal();
                        break;
					
                }
            }
        }
		
        //
        // Communication with server
        // 
        protected internal virtual void  write(byte[] buf, int size)
        {
            lock(cs)
            {
                try
                {
                    socket.Send(buf, 0, size, SocketFlags.None);
                }
                catch (System.IO.IOException x)
                {
                    database.handleException(x);
                }
            }
        }
		
        internal void  write(byte[] buf)
        {
            write(buf, buf.Length);
        }
		
		
        protected internal virtual void  read(byte[] buf, int size)
        {
            try
            {
                int offs = 0;
                while (size > 0)
                {
                    int len = socket.Receive(buf, offs, size, SocketFlags.None);
                    if (len < 0)
                    {
                        database.handleError("Broken socket");
                    }
                    offs += len;
                    size -= len;
                }
            }
            catch (System.IO.IOException x)
            {
                database.handleException(x);
            }
        }
		
        protected internal void  sendReceive(byte[] req, int expectedCmd)
        {
            sendReceive(req, req.Length, expectedCmd);
        }
		
        protected internal virtual void  sendReceive(byte[] req, int len, int expectedCmd)
        {
            lock (requestMutex)
            {
                if (req != null)
                {
                    write(req, len);
                }
                repEvent.waitSignal();
                System.Diagnostics.Debug.Assert(reply[Protocol.req_cmd_offs] == expectedCmd);
                Array.Copy(reply, 0, result, 0, Protocol.req_size);
            }
            rcvEvent.signal();
        }
		
        protected internal byte[] sendReceiveBody(byte[] req, int expectedCmd)
        {
            return sendReceiveBody(req, req.Length, expectedCmd);
        }
		
        protected internal virtual byte[] sendReceiveBody(byte[] req, int len, int expectedCmd)
        {
            byte[] body;
            lock (requestMutex)
            {
                if (req != null)
                {
                    write(req, len);
                }
                repEvent.waitSignal();
                System.Diagnostics.Debug.Assert(reply[Protocol.req_cmd_offs] == expectedCmd);
                Array.Copy(reply, 0, result, 0, Protocol.req_size);
    			
                int bodyLen = Protocol.unpack4(reply, Protocol.req_size_offs);
                System.Diagnostics.Debug.Assert(bodyLen > 0);
                body = new byte[bodyLen];
                read(body, bodyLen);
            }    			
            rcvEvent.signal();
            return body;
        }
		
        protected internal virtual bool lockObject(ObjectHeader obj, int lockType, int lockAttr)
        {
            lock(this)
            {
                request[Protocol.req_cmd_offs] = Protocol.cmd_lock;
                request[Protocol.req_type_offs] = (byte) lockType;
                Protocol.pack2(request, Protocol.req_attr_offs, lockAttr);
                Protocol.pack4(request, Protocol.req_opid_offs, obj.opid);
                sendReceive(request, Protocol.cmd_lockresult);
                return result[Protocol.req_status_offs] != 0;
            }
        }
		
        protected internal virtual void  unlock(ObjectHeader obj, int lockType)
        {
            lock(this)
            {
                request[Protocol.req_cmd_offs] = Protocol.cmd_unlock;
                request[Protocol.req_type_offs] = (byte) lockType;
                Protocol.pack4(request, Protocol.req_opid_offs, obj.opid);
                write(request);
            }
        }
		
        private void  bulkAllocate(int nReserved)
        {
            int i;
            AllocBuf[] buf = bulkAllocBuf;
            int nAllocated = allocBufPos;
            allocBufPos = 0;
            byte[] req = new byte[Protocol.req_size + nAllocated * 6];
            req[Protocol.req_cmd_offs] = Protocol.cmd_bulk_alloc;
            Protocol.pack4(req, Protocol.req_size_offs, nAllocated);
            Protocol.pack4(req, Protocol.req_reserve_offs, nReserved);
            int pos = Protocol.req_size;
            for (i = 0; i < nAllocated; i++)
            {
                pos = Protocol.pack4(req, pos, buf[i].size);
            }
            for (i = 0; i < nAllocated; i++)
            {
                pos = Protocol.pack2(req, pos, buf[i].cpid);
            }
            byte[] reply = sendReceiveBody(req, Protocol.cmd_opids);
            pos = 0;
            for (i = 0; i < nReserved; i++)
            {
                buf[i].opid = Protocol.unpack4(reply, i * 4);
            }
        }
		
        protected internal virtual void  allocate(ObjectHeader obj, int align)
        {
            lock(this)
            {
                ClassDescriptor desc = obj.desc;
                int cpid = 0;
                if (desc.cid < cpidTableSize)
                {
                    cpid = cpidTable[desc.cid];
                }
                if (cpid == 0)
                {
                    cpid = registerClass(desc.dbsClass);
                    System.Diagnostics.Debug.Assert(cpid != 0);
                    associateCpidWithClass(desc, cpid);
                }
                int bufSize = database.allocBufSize;
                if (bufSize != 0)
                {
                    if (allocBufSize == 0)
                    {
                        bulkAllocBuf = new AllocBuf[bufSize];
                        for (int i = bufSize; --i >= 0; bulkAllocBuf[i] = new AllocBuf())
                            ;
                        allocBufPos = 0;
                        allocBufSize = bufSize;
                        bulkAllocate(bufSize);
                    }
                    else if (allocBufPos == allocBufSize)
                    {
                        bulkAllocate(allocBufSize);
                    }
                    AllocBuf buf = bulkAllocBuf[allocBufPos++];
                    buf.cpid = cpid;
                    buf.size = obj.size();
                    if (align != 0)
                    {
                        buf.size |= Protocol.ALLOC_ALIGNED;
                    }
                    obj.opid = buf.opid;
                }
                else
                {
                    request[Protocol.req_cmd_offs] = Protocol.cmd_alloc;
                    request[Protocol.req_align_offs] = (byte) align;
                    Protocol.pack2(request, Protocol.req_cpid_offs, cpid);
                    Protocol.pack4(request, Protocol.req_size_offs, obj.size());
                    sendReceive(request, Protocol.cmd_location);
                    obj.opid = Protocol.unpack4(result, Protocol.req_opid_offs);
                }
                System.Diagnostics.Debug.Assert(obj.opid != 0);
                objectCache.put(obj.opid, obj);
            }
        }
		
        protected internal virtual void  deallocate(ObjectHeader obj)
        {
            lock(this)
            {
                request[Protocol.req_cmd_offs] = Protocol.cmd_alloc;
                Protocol.pack4(request, Protocol.req_opid_offs, obj.opid);
                write(request);
            }
        }
		
        protected internal virtual int registerClass(byte[] desc)
        {
            lock(this)
            {
                byte[] req = new byte[Protocol.req_size + desc.Length];
                Array.Copy(desc, 0, req, Protocol.req_size, desc.Length);
                req[Protocol.req_cmd_offs] = Protocol.cmd_putclass;
                Protocol.pack4(req, Protocol.req_size_offs, desc.Length);
                sendReceive(req, Protocol.cmd_classid);
                return Protocol.unpack2(result, Protocol.req_cpid_offs);
            }
        }
		
        protected internal virtual byte[] getClass(int cpid)
        {
            lock(this)
            {
                request[Protocol.req_cmd_offs] = Protocol.cmd_getclass;
                Protocol.pack2(request, Protocol.req_cpid_offs, cpid);
                return sendReceiveBody(request, Protocol.cmd_classdesc);
            }
        }
		
        protected internal virtual void  changeClass(int cpid, byte[] desc)
        {
            lock(this)
            {
                byte[] req = new byte[Protocol.req_size + desc.Length];
                Array.Copy(desc, 0, req, Protocol.req_size, desc.Length);
                req[Protocol.req_cmd_offs] = Protocol.cmd_putclass;
                Protocol.pack2(request, Protocol.req_cpid_offs, cpid);
                Protocol.pack4(req, Protocol.req_size_offs, desc.Length);
                write(req);
            }
        }
		
        protected internal virtual void  throwObject(ObjectHeader obj)
        {
            if (closing)
            {
                return ;
            }
            lock(notification)
            {
                notification[Protocol.req_cmd_offs] = Protocol.cmd_throw;
                Protocol.pack4(notification, Protocol.req_opid_offs, obj.opid);
                Protocol.pack4(notification, Protocol.req_extra_offs, 0);
                write(notification);
            }
        }
		
        protected internal virtual void  forgetObject(ObjectHeader obj)
        {
            if (closing)
            {
                return ;
            }
            lock(notification)
            {
                objectCache.remove(obj.opid);
                notification[Protocol.req_cmd_offs] = Protocol.cmd_forget;
                Protocol.pack4(notification, Protocol.req_opid_offs, obj.opid);
                Protocol.pack4(notification, Protocol.req_extra_offs, 0);
                write(notification);
            }
        }
		
        protected internal virtual void  forgetObjects(int[] opids)
        {
            if (closing)
            {
                return ;
            }
            byte[] notifications = new byte[opids.Length * Protocol.req_size];
            int offs = 0;
            for (int i = opids.Length; --i >= 0; )
            {
                notifications[Protocol.req_cmd_offs + offs] = Protocol.cmd_forget;
                Protocol.pack4(notifications, Protocol.req_opid_offs + offs, opids[i]);
                offs += Protocol.req_size;
            }
            Protocol.pack4(notifications, Protocol.req_extra_offs, opids.Length - 1);
            write(notifications, offs);
        }
		
        protected internal virtual int commitTransaction(int nTransServers, Storage[] servers)
        {
            lock(this)
            {
                if (allocBufPos != 0)
                {
                    bulkAllocate(allocBufPos);
                }
                int transSize = transBufferOffs;
                if (nTransServers > 1)
                {
                    appendTransBuffer(nTransServers * 2);
                    for (int i = 0; i < nTransServers; i++)
                    {
                        transSize = Protocol.pack2(transBuffer, transSize, servers[i].id);
                    }
                }
                transBuffer[Protocol.req_cmd_offs] = Protocol.cmd_transaction;
                transBuffer[Protocol.req_n_servers_offs] = (byte) nTransServers;
                Protocol.pack4(transBuffer, Protocol.req_size_offs, transSize - Protocol.req_size);
                sendReceive(transBuffer, transSize, Protocol.cmd_transresult);
                if (transSize > maxTransBufferSize)
                {
                    transBuffer = null;
                }
                transBufferOffs = 0;
                return result[Protocol.req_status_offs] != 0?Protocol.unpack4(result, Protocol.req_tid_offs):- 1;
            }
        }
		
        protected internal virtual void  commitSubtransaction(Storage coordinator, int nTransServers, Storage[] servers, int tid)
        {
            lock(this)
            {
                if (allocBufPos != 0)
                {
                    bulkAllocate(allocBufPos);
                }
                int transSize = transBufferOffs;
                if (nTransServers > 1)
                {
                    appendTransBuffer(nTransServers * 2);
                    for (int i = 0; i < nTransServers; i++)
                    {
                        transSize = Protocol.pack2(transBuffer, transSize, servers[i].id);
                    }
                }
                transBuffer[Protocol.req_cmd_offs] = Protocol.cmd_subtransact;
                transBuffer[Protocol.req_n_servers_offs] = (byte) nTransServers;
                Protocol.pack2(transBuffer, Protocol.req_coordinator_offs, coordinator.id);
                Protocol.pack4(transBuffer, Protocol.req_size_offs, transSize - Protocol.req_size);
                Protocol.pack4(transBuffer, Protocol.req_tid_offs, tid);
                write(transBuffer, transSize);
                if (transSize > maxTransBufferSize)
                {
                    transBuffer = null;
                }
                transBufferOffs = 0;
            }
        }
		
        protected internal virtual bool waitTransactionCompletion()
        {
            lock(this)
            {
                sendReceive(null, 0, Protocol.cmd_transresult);
                return result[Protocol.req_status_offs] != 0;
            }
        }
		
        protected internal virtual void  open(String host, int port)
        {
            lock(this)
            {
                createWeakObjectCache();
                socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                foreach (IPAddress ip in Dns.Resolve(host).AddressList) 
                { 
                    try 
                    {
                        socket.Connect(new IPEndPoint(ip, port));	
                        break;
                    } 
                    catch (SocketException) {}
                }
                socket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, 1);
                socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.Linger,
                    new System.Net.Sockets.LingerOption(true, LingerTime));
 
                String hostName = System.Net.Dns.GetHostName();
                long time = DateTime.Now.Ticks / 10000;
                String clientId = hostName + "-" + System.Threading.Thread.CurrentThread.GetHashCode() + "(" + time / (1000 * 60 * 60) % 24 + ":" + time / (1000 * 60) % 60 + ":" + time / 1000 % 60 + "." + time % 1000 + ")";
                int nameLen = clientId.Length;
                byte[] req = new byte[Protocol.req_size + nameLen + 1];
                req[Protocol.req_cmd_offs] = Protocol.cmd_login;
                Protocol.pack4(req, Protocol.req_size_offs, nameLen + 1);
                Protocol.packAscii(req, Protocol.req_size, clientId);
                write(req);
                read(reply, Protocol.req_size);
                switch (reply[Protocol.req_cmd_offs])
                {
                    case Protocol.cmd_bye: 
                        database.disconnected(id);
                        goto case Protocol.cmd_refused;
					
                    case Protocol.cmd_refused: 
                        throw new ConnectionRefusedException();
					
                    default: 
                        System.Diagnostics.Debug.Assert(reply[Protocol.req_cmd_offs] == Protocol.cmd_ok);
                        break;
					
                }
				
                closing = false;
                receiver = new System.Threading.Thread(new System.Threading.ThreadStart(Run));
                receiver.IsBackground = true;
                receiver.Start();
            }
        }
		
        protected internal virtual void  close()
        {
            lock(this)
            {
                closing = true;
                request[Protocol.req_cmd_offs] = Protocol.cmd_logout;
                sendReceive(request, Protocol.cmd_bye);
                receiver.Join();
				
                try
                {
                    socket.Close();
                }
                catch (System.IO.IOException x)
                {
                    database.handleException(x);
                }
            }
        }
		
        		
        protected internal Storage(Database db, int sid)
        {
            smallReqbuf = new byte[MAX_REFS * Protocol.req_size];
            smallPermutation = new int[MAX_REFS];
            smallSids = new int[MAX_REFS];
            smallRefs = new ObjectHeader[MAX_REFS];
            database = db;
            id = sid;
            classDictionarySize = 1024;
            classDictionary = new ClassDescriptor[classDictionarySize];
            cpidTableSize = 1024;
            cpidTable = new int[cpidTableSize];
            reply = new byte[Protocol.req_size];
            result = new byte[Protocol.req_size];
            request = new byte[Protocol.req_size];
            notification = new byte[Protocol.req_size];
            rcvEvent = new CondEvent();
            repEvent = new CondEvent();
            loadList = new ObjectHeader(null);
            cs = new Object();
    	    requestMutex = new Object();
        }
    }
}