//-< TESTTRANS.CXX >-------------------------------------------------*--------*
// GOODS                     Version 2.37        (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     15-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 15-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test of per-thread transactions
//-------------------------------------------------------------------*--------*

#include "goods.h"

USE_GOODS_NAMESPACE

const int nNodes = 100;
const int nIterations = 100;
const int nThreadsByDefault = 10;

static semaphore done;

class L2ListNode: public object { 
  public:
    METACLASS_DECLARATIONS(L2ListNode, object);

    ref<L2ListNode> next;
    ref<L2ListNode> prev;
    int             value;

    L2ListNode() : object(self_class) {}

    void buildList() { 
        if (next.is_nil()) { 
            next = this;
            prev = this;
            value = 0;
            for (int i = 1; i <= nNodes; i++) {
                L2ListNode* node = new L2ListNode();
                node->value = i;
                link(node);
            }
        }
    }

    void increment() { 
        value += 1;
    }

    void decrement() { 
        value -= 1; 
    }

    void update() { 
        increment();
        modify(next)->decrement();
    }

    void link(L2ListNode* node) {
        node->next = this;
        node->prev = prev;
        modify(prev)->next = node;
        prev = node;
    }

    int calculateSum() const { 
        int sum = 0;
        for (ref<L2ListNode> node = next; node != this; node = node->next) {
            sum += node->value;
        }
        return sum;
    }
    
    void updateNode(int index) const { 
        ref<L2ListNode> node = next; 
        while (--index >= 0) { 
            node = node->next;
        }
        modify(node)->update();
    }   

    void reset() { 
        next = NULL;
        prev = NULL;
    }

    void initialize() const { 
        if (is_abstract_root()) { 
            ref<L2ListNode> root = this;
            modify(root)->become(new L2ListNode());
        }
    }
};

field_descriptor& L2ListNode::describe_components()
{
    return FIELD(next), FIELD(prev), FIELD(value);
}


REGISTER(L2ListNode, object, pessimistic_repeatable_read_scheme);

void task_proc run(void*) { 
    database db;
    db.attach();
    if (db.open("testtrans.cfg")) {
        ref<L2ListNode> list;
        db.get_root(list);
        list->initialize();
        modify(list)->buildList();
        ref<L2ListNode> node;
        for (int i = 0; i < nIterations; i++) { 
            int sum = list->calculateSum();
            if (sum != nNodes*(nNodes+1)/2) { 
                fprintf(stderr, "Wrong check sum: %d instead of %d\n",   
                        sum, nNodes*(nNodes+1)/2);
            }
            list->updateNode(i % (nNodes-1));
        }
        db.close();
    } else { 
        fprintf(stderr, "Thread failed to open database\n");
    }
    db.detach();
    done.signal();
}
 
int main(int argc, char* argv[])
{
    int nThreads = nThreadsByDefault;
    int i;
    task::initialize(task::normal_stack);
    if (argc > 1) { 
        nThreads = atoi(argv[1]);
    }
    for (i = 0; i < nThreads; i++) { 
        task::create(run);
    }
    for (i = 0; i < nThreads; i++) { 
        done.wait();
    }
    database db;
    if (db.open("testtrans.cfg")) {
        ref<L2ListNode> list;
        db.get_root(list);
        if (list->next->value != 1 + (nIterations+nNodes-2)/(nNodes-1)*nThreads 
            || list->prev->value != nNodes - nIterations/(nNodes-1)*nThreads) 
        { 
            fprintf(stderr, "Something is definitly wrong: list->next->value=%d instead of %d, "
                    "list->prev->value=%d instead of %d\n", 
                    list->next->value, 1 + (nIterations+nNodes-2)/(nNodes-1)*nThreads,  
                    list->prev->value, nNodes - nIterations/(nNodes-1)*nThreads);
        } else { 
            printf("Test successfully passed\n");
        }       
        modify(list)->reset();
        db.close();
        return EXIT_SUCCESS;
    } else {
        fprintf(stderr, "Failed to open detabase\n");
        return EXIT_FAILURE;
    }
}





