//-< TestBulkAlloc.java >------------------------------------------------*--------*
// GOODS                     Version 2.36        (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     25-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 25-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test of bulk alloc mode
//-------------------------------------------------------------------*--------*

import goodsjpi.*;
import goodslib.*;
import java.util.Random;

class Component extends Persistent { 
    Component next;
    int       value;

    Component(int index, Component chain) { 
	next = chain;
	value = index;
    } 

    static final Metaobject defaultMetaobject = new OptimisticMetaobject();
};

class BtreeRecord extends SetMember { 
    Component  components;
    long       key;

    public int compare(Ordered mbr) { 
	long anotherKey = ((BtreeRecord)mbr).key;
	return (key < anotherKey) ? -1 : (key == anotherKey) ? 0 : 1;
    }
	
    public int compareKey(Object key) { 
	long key1 = this.key;
	long key2 = ((Long)key).longValue();
	return (key1 < key2) ? -1 : (key1 == key2) ? 0 : 1;
    }

    BtreeRecord(long key) { 
	this.key = key;
    }

    static final Metaobject defaultMetaobject = new OptimisticMetaobject();
};


public class TestBulkAlloc extends Btree { 
    static final int nComponents = 10;
    static final int bTreePageSize = 128;
    static final int nRecords = 10000;

    void insert(long key) { 
	BtreeRecord rec = new BtreeRecord(key);
	for (int i = nComponents; --i >= 0;) { 
	    rec.components = new Component(i, rec.components);
	}
	super.insert(rec);
    }

    void locate(long key) { 
	SetMember mbr = find(new Long(key));
	int sum = 0;
	for (Component cmp = ((BtreeRecord)mbr).components; cmp != null; cmp = cmp.next) {
	    sum += cmp.value;
	}
	Assert.that(sum == nComponents*(nComponents-1)/2);
    }
    
    TestBulkAlloc() { 
	super(bTreePageSize);
    }
    
    static public void main(String args[]) { 
	Database db = new Database();
	if (db.open("testtrans.cfg")) {
	    db.setAllocBufferSize(100);
	    TestBulkAlloc root = (TestBulkAlloc)db.getRoot();
	    if (root == null) { 
		root = new TestBulkAlloc();
		db.setRoot(root);
	    }
	    Random rnd = new Random(2001);
	    int i, j, step;
	    long start = System.currentTimeMillis();
	    long last = start, curr;
	    for (i = 0, j = 0, step = 100; i < nRecords; i++) { 
		root.insert(rnd.nextLong());
		if (++j == step) {  
		    curr = System.currentTimeMillis();
		    if (curr != last) { 
			System.out.print("Inserted " + (i+1) + " records, " 
					 + ((i+1)*1000 / (curr - start))
					 + " records per second (" 
					 + (j*1000 / (curr - last))
					 + " for last " + j  + ")    \r");
			j = 0;
			last = curr;
		    } else { 
			step *= 10;
		    }
		}
	    }
	    curr = System.currentTimeMillis();
	    System.out.println("\nEllapsed time " + (curr-start) + " for " + nRecords 
			       + " records insertion (" 
			       + (nRecords*1000/(curr-start)) + " records per second)");
	    rnd.setSeed(2001);
	    start = System.currentTimeMillis();
	    for (i = 0; i < nRecords; i++) { 
		root.locate(rnd.nextLong());
	    }
	    curr = System.currentTimeMillis();
	    System.out.println("Ellapsed time " + (curr-start) + " for locating " 
			       + nRecords + " records (" 
			       + (nRecords*1000 / (curr-start)) + " per second)"); 
	    db.close();
	} else { 
	    System.err.println("Failed to open database");
	}
    }
}





